# Create all swatches for public domain AKON

import pandas as pd
import requests
import jsonpath_ng as jp

from PIL import Image
from io import BytesIO
from colorsys import rgb_to_hsv, hsv_to_rgb
from scipy.cluster.vq import kmeans
from numpy import array


DEFAULT_NUM_COLORS = 6
SCALE = 256.0

DEFAULT_MINV = 0
DEFAULT_MAXV = 256


def get_raw_data():
    df = pd.read_csv(filepath_or_buffer='https://labs.onb.ac.at/gitlab/labs-team/raw-metadata/raw/master/akon_postcards_public_domain.csv.bz2', compression='bz2')
    out_df = df[['akon_id']]
    return out_df


jp_image_link = jp.parse('sequences[*].canvases[*].images[*].resource.@id')

def get_image_link_for_akon_id(akon_id):
    r = requests.get(f'https://iiif.onb.ac.at/presentation/AKON/{akon_id}/manifest')
    r.raise_for_status()
    links = [m.value for m in jp_image_link.find(r.json())]
    if len(links) > 1:
        print(f'{len(links)} images found for id {akon_id}.', file=sys.stderr)
    return links[0]


def get_clamped_image_link(original_iiif_image_link, max_width=200, max_height=200):
    slices = original_iiif_image_link.split('/')
    slices[-3] = f'!{max_width:d},{max_height:d}'
    return '/'.join(slices)


def down_scale(x):
    return x / SCALE


def up_scale(x):
    return int(x * SCALE)


def clamp_hsv(color, min_v, max_v):
    """
    Clamps a color such that the value (lightness) is between min_v and max_v.
    """
    # use down_scale to convert color to value between 0-1 as expected by rgb_hsv
    h, s, v = [down_scale(c) for c in color]
    # also convert the min_v and max_v to values between 0-1
    min_v, max_v = map(down_scale, (min_v, max_v))
    # get the maximum of the min value and the color's value (therefore bumping it up if needed)
    # then get the minimum of that number and the max_v (bumping the value down if needed)
    v = min(max(min_v, v), max_v)
    # apply upscale to get the h, s, v(which has been clamped) back to 0-255, return as tuple
    return tuple(map(up_scale, (h, s, v)))


def clip_hsv(colors_hsv, min_v):
    min_v = down_scale(min_v)
    hsvs = [tuple(map(down_scale, color)) for color in colors_hsv]
    hsvs = filter(lambda hsv: hsv[2] >= min_v, hsvs)
    return [tuple(map(up_scale, hsv)) for hsv in hsvs]


def order_by_hue_hsv(colors):
    """
    Orders colors by hue.
    """
    hsvs = [list(map(down_scale, color)) for color in colors]
    hsvs.sort(key=lambda t: t[0])
    return [tuple(map(up_scale, hsv_to_rgb(*hsv))) for hsv in hsvs]


def get_colors(img):
    """
    Returns a list of all the image's colors.
    """
    w, h = img.size
    # convert('HSV') converts the image's pixels info to HSV
    # getcolors() returns an unsorted list of (count, pixel) values
    # w * h ensures that maxcolors parameter is set so that each pixel could be unique
    # there are three values returned in a list
    return [single_color for count, color in img.convert('HSV').getcolors(w * h) for single_color in [color] * count]


def hexify(rgb):
    return "#{0:02x}{1:02x}{2:02x}".format(*rgb)


def colorz(image_url, n=DEFAULT_NUM_COLORS, min_v=DEFAULT_MINV, max_v=DEFAULT_MAXV,
           order_colors=True, coefficients=[1.0, 1.0, 1.0], clip_colors=False, min_clip_s=20, min_clip_v=20):
    """
    Get the n most dominant colors of an image.

    Total number of colors returned is n, optionally ordered by hue.
    Returns as a list of RGB triples.

    """
    r = requests.get(image_url)
    r.raise_for_status()
    img = Image.open(BytesIO(r.content))
    
    obs = get_colors(img) # gets a list of HSV colors (e.g. (213, 191, 152)) for each pixel
    # adjust the value of each color, if you've chosen to change minimum and maximum values
    clamped = [clamp_hsv(color, min_v, max_v) for color in obs]
    clipped = clip_hsv(clamped, min_clip_v) if clip_colors else clamped
    # turns the list of colors into a numpy array of floats, then applies scipy's k-means function
    clusters, _ = kmeans(array(clipped).astype(float) * coefficients, n)
    normalized_clusters = clusters / coefficients
    colors = order_by_hue_hsv(normalized_clusters) if order_colors else normalized_clusters
    
    hex_colors = list(map(hexify, colors)) # turn RGB into hex colors for web
    return hex_colors


def draw_row_with_links(link_and_colors):
    html = ""
    url = link_and_colors[0]
    for count, color in enumerate(link_and_colors[1]):
        square = '<rect x="{0}" y="{1}" width="30" height="30" fill="{2}" />'.format(((count * 30)), 0, color)
        html += square
    full_html = '<a href="{0}" target="_blank"><svg height="30" width="{2}">{1}</svg>'.format(url, html, len(link_and_colors) * 30)
    return full_html


def get_link_and_colors(akon_id):
    image_link = get_image_link_for_akon_id(akon_id)
    thumb_link = get_clamped_image_link(image_link)
    color_swatches = colorz(thumb_link, 6, True)
    return image_link, color_swatches


def get_color_swatches(image_link):
    print(image_link, end='')
    thumb_link = get_clamped_image_link(image_link)
    print('.', end='')
    color_swatches = colorz(thumb_link, 6, order_colors=True, coefficients=[1.0, 2.0, 0.6], clip_colors=True, min_clip_s=20, min_clip_v=20)
    print('.')
    return color_swatches


if __name__ == '__main__':
    FILE_NAME = 'historic_postcards_color_swatches.csv.bz2'
    
    df = get_raw_data()
    print('Data loaded.')
    # For testing, use only a tiny part of the available data
    # df = df.sample(100).copy()
    df['image_link'] = df['akon_id'].apply(get_image_link_for_akon_id)
    print('Image links created.')
    df['hex_colors'] = df['image_link'].apply(get_color_swatches)
    print('Colors computed.')
    df['html'] = df[['image_link', 'hex_colors']].apply(draw_row_with_links, axis='columns')
    print('HTML created.')
    df.to_csv(FILE_NAME, compression='bz2')
    print(f'Output file {FILE_NAME} written.')

